from .models import RootCA, IntermediateCA, KeyValueEngine
from .vault import VaultClient
from .cli import cli
from . import schemas, utils
from distutils.util import strtobool
import os.path
import requests
import sys


def main():
    parser = cli()
    args = parser.parse_args()

    if args.subcommand is None:
        parser.print_help()
        sys.exit()

    if args.baseurl is None:
        args.baseurl = utils.get_from_environment('VAULT_ADDR')

    if args.tls_skip_verify is None:
        args.tls_skip_verify = strtobool(utils.get_from_environment('VAULT_SKIP_VERIFY'))

    verify_ssl = True
    if args.tls_skip_verify:
        verify_ssl = False
        requests.packages.urllib3.disable_warnings()

    if args.subcommand == 'init':
        vault_client = VaultClient(baseurl=args.baseurl, verify_ssl=verify_ssl)

        initialized, sealed = vault_client.healthcheck()

        if not initialized:
            vault_client.initialize_server()

        if sealed:
            vault_client.unseal_server()

    elif args.subcommand == 'apply':
        # authentication token is required to talk to Vault
        vault_token = utils.get_from_environment('VAULT_TOKEN')

        vault_client = VaultClient(baseurl=args.baseurl, token=vault_token, debugging=args.debugging, verify_ssl=verify_ssl)

        path = os.path.abspath(os.path.expanduser(args.file))

        manifest_files = []
        documents      = []

        if os.path.isdir(path):
            manifest_files.extend(utils.get_manifest_files(path))
        else:
            manifest_files.append(path)

        for filepath in manifest_files:
            documents.extend(utils.read_manifest_file(filepath))

        _, sealed = vault_client.healthcheck()
        if sealed:
            sys.exit(1)

        roots, intermediates, kv_engines = utils.get_validated_manifests(documents)

        # mount KV engines
        for kve in kv_engines:
            kvengine = KeyValueEngine(args.baseurl, kve)
            vault_client.mount_kv_engine(kvengine)

        # create the Root CAs
        for ca in roots:
            manifest = schemas.RootCASchema(ca)
            root_ca  = RootCA(args.baseurl, manifest)

            vault_client.mount_pki_engine(root_ca)
            vault_client.create_root_ca(root_ca)
            if not vault_client.check_existing_ca(root_ca, quiet=True):
                vault_client.configure_ca_urls(root_ca)

        intermediates = utils.sort_intermediate_certificate_authorities(intermediates)

        # create the Intermediate CAs
        for ca in intermediates:
            manifest        = schemas.IntermediateCASchema(ca)
            intermediate_ca = IntermediateCA(args.baseurl, manifest)

            vault_client.mount_pki_engine(intermediate_ca)

            if not vault_client.check_existing_ca(intermediate_ca):
                vault_client.create_intermediate_ca(intermediate_ca)
                vault_client.sign_intermediate_ca(intermediate_ca)
                vault_client.set_intermediate_ca(intermediate_ca)
                vault_client.configure_ca_urls(intermediate_ca)
                vault_client.set_crl_configuration(intermediate_ca)

                if intermediate_ca.catype == 'exported':
                    vault_client.store_ca_private_key(intermediate_ca)

            vault_client.configure_ca_roles(intermediate_ca)
            vault_client.configure_ca_policies(intermediate_ca)



from voluptuous import Schema, Required, Optional, All, Any, Range, Match, Coerce

MOUNT_PATH_REGEX = r'^(?![-\/])[a-z0-9-_\/]+(?<![-\/])$'

RootCASchema = Schema({
    Required('kind'): All('RootCA', msg="Must be 'RootCA'"),
    Required('metadata'): {
        Required('name'): Match(MOUNT_PATH_REGEX, msg="Must be lowercase alphanumberic string"),
        Required('description'): str
    },
    Required('spec'): {
        Required('key_type'): Any('rsa', 'ec', msg="Must be 'rsa' or 'ec'"),
        Required('key_bits'): Range(min=256, max=4096),
        Optional('ttl', default='87660h'): Match(r'\d+h'),
        Optional('exclude_cn_from_sans', default=True): bool,
        Required('subject'): {
            Required('common_name'): str,
            Optional('country'): Match(r'[A-Z]{2}'),
            Optional('locality'): str,
            Optional('province'): str,
            Optional('organization'): str,
            Optional('ou'): str
        }
    }
})

IntermediateCASchema = Schema({
    Required('kind'): All('IntermediateCA', msg="Must be 'IntermediateCA'"),
    Required('metadata'): {
        Required('name'): Match(MOUNT_PATH_REGEX, msg="Must be lowercase alphanumberic string"),
        Required('description'): str,
        Required('issuer'): Match(MOUNT_PATH_REGEX, msg="Must be lowercase alphanumberic string"),
        Optional('kv_engine'): Match(MOUNT_PATH_REGEX, msg="Must be lowercase alphanumberic string")
    },
    Required('spec'): {
        Required('type'): Any('internal', 'exported', msg="Must be 'internal' or 'exported'"),
        Required('key_type'): Any('rsa', 'ec', msg="Must be 'rsa' or 'ec'"),
        Required('key_bits'): Range(min=256, max=4096),
        Optional('ttl', default='87660h'): Match(r'\d+h'),
        Optional('exclude_cn_from_sans', default=True): bool,
        Optional('max_path_length', default=0): Range(min=-1, max=5),
        Optional('crl', default={}): {
            Optional('expiry', default='72h'): Match(r'\d+[hms]'),
            Optional('disable', default=False): bool
        },
        Required('subject'): {
            Required('common_name'): str,
            Optional('country'): Match(r'[A-Z]{2}'),
            Optional('locality'): str,
            Optional('province'): str,
            Optional('organization'): str,
            Optional('ou'): str,
        },
        Optional('roles', default=[]): [{
            Required('name'): Match(r'^[a-z0-9-_]+$', msg="Must be lowercase alphanumberic string"),
            Required('config'): {
                Required('max_ttl'): Match(r'\d+[hms]'),
                Optional('ttl'): Match(r'\d+[hms]'),
                Required('server_flag'): bool,
                Required('client_flag'): bool,
                Optional('allow_localhost'): bool,
                Optional('allow_subdomains'): bool,
                Optional('allow_any_name'): bool,
                Optional('allow_ip_sans'): bool,
                Optional('enforce_hostnames'): bool,
                Optional('generate_lease'): bool,
                Optional('no_store'): bool,
                Optional('allowed_domains'): [Match(r'^(?![-.])[a-zA-Z0-9-\.]+(?<![.-])$')],
            }
        }],
        Optional('policies', default=[]): [{
            Required('name'): Match(MOUNT_PATH_REGEX, msg="Must be lowercase alphanumberic string"),
            Required('policy'): str
        }]
    }
})

KeyValueSchema = Schema({
    Required('kind'): All('KV', msg="Must be 'KV'"),
    Required('metadata'): {
        Required('name'): Match(MOUNT_PATH_REGEX, msg="Must be lowercase alphanumberic string"),
        Required('description'): str
    },
    Required('spec'): {
        Optional("config"): {
            Optional('default_lease_ttl', default='8766h'): Match(r'\d+[hms]'),
            Optional('max_lease_ttl', default='17532h'): Match(r'\d+[hms]'),
            Optional('force_no_cache', default=False): bool
        },
        Required("options"): {
            Required("version"): All(Any(1, "1", msg="Must be '1'"), Coerce(str))
        }
    }
})



from urllib.parse import urljoin


class CertificateAuthority:
    def __init__(self, baseurl, manifest):
        self.baseurl = baseurl
        self.dict = manifest
        self.name = manifest['metadata']['name']
        self.description = manifest['metadata']['description']

    @property
    def spec(self):
        spec = self.dict['spec'].copy()

        subject = spec.get('subject')
        if subject:
            spec.pop('subject')
            spec.update(subject)
        return spec

    @property
    def backend(self):
        return {'type': 'pki', 'description': self.description, 'config': {'max_lease_ttl': self.ttl}}

    @property
    def ttl(self):
        return self.spec['ttl']

    @property
    def config_url(self):
        return urljoin(self.baseurl, f"/v1/{self.name}/config/urls")

    @property
    def ca_urls(self):
        return {
            'issuing_certificates': f'{self.baseurl}/v1/{self.name}/ca',
            'crl_distribution_points': f'{self.baseurl}/v1/{self.name}/crl',
        }


class RootCA(CertificateAuthority):
    def __init__(self, baseurl, manifest):
        super().__init__(baseurl, manifest)

    @property
    def url(self):
        return urljoin(self.baseurl, f"/v1/{self.name}/root/generate/internal")


class IntermediateCA(CertificateAuthority):
    def __init__(self, baseurl, manifest):
        super().__init__(baseurl, manifest)
        self._csr = None
        self._private_key = None
        self._cert = None

    @property
    def issuer(self):
        return self.dict['metadata']['issuer']

    @property
    def kv_engine(self):
        return self.dict['metadata']['kv_engine']

    @property
    def catype(self):
        return self.spec['type']

    @property
    def csr(self):
        return self._csr

    @csr.setter
    def csr(self, value):
        self._csr = value

    @property
    def cert(self):
        return self._cert

    @cert.setter
    def cert(self, value):
        self._cert = value

    @property
    def private_key(self):
        return {'private_key': self._private_key}

    @private_key.setter
    def private_key(self, value):
        self._private_key = value

    @property
    def url(self):
        return urljoin(self.baseurl, f"/v1/{self.name}/intermediate/generate/{self.catype}")

    @property
    def crl_config_url(self):
        return urljoin(self.baseurl, f"/v1/{self.name}/config/crl")

    @property
    def issuer_sign_url(self):
        return urljoin(self.baseurl, f"/v1/{self.issuer}/root/sign-intermediate")

    @property
    def set_signed_url(self):
        return urljoin(self.baseurl, f"/v1/{self.name}/intermediate/set-signed")

    @property
    def kv_engine_url(self):
        return urljoin(self.baseurl, f"/v1/{self.kv_engine}/{self.name}")

    @property
    def spec(self):
        spec = self.dict['spec'].copy()

        subject = spec.get('subject')
        if subject:
            spec.pop('subject')
            spec.update(subject)
        if spec.get('crl'):
            spec.pop('crl')
        if spec.get('roles'):
            spec.pop('roles')
        if spec.get('policies'):
            spec.pop('policies')
        return spec

    @property
    def crl_config(self):
        return self.dict['spec']['crl']

    @property
    def roles(self):
        return self.dict['spec']['roles']

    @property
    def policies(self):
        return self.dict['spec']['policies']


class KeyValueEngine:
    def __init__(self, baseurl, manifest):
        self.baseurl = baseurl
        self.dict = manifest
        self.name = manifest['metadata']['name']

    @property
    def spec(self):
        spec = self.dict['spec'].copy()
        spec.update({
            'type': self.dict['kind'].lower(),
            'description': self.dict['metadata']['description']
        })
        return spec

    @property
    def url(self):
        return urljoin(self.baseurl, f"v1/sys/mounts/{self.name}")
